"""2-of-2 example"""


from musig import (
    PublicKey,
    PartialSignature,
    SigningSession,
    ProtocolState,
    ProtocolMessage,
    ProtocolError,
    Nonce,
    NonceCommitment
)
from nacl.signing import SignedMessage, SigningKey, VerifyKey
from secrets import token_bytes


# simulate sending message
def send(pm: ProtocolMessage):
    print(f'message: {repr(pm)}')
    print(f'sending: {pm}\n')


# simulate receiving a message
def receive(pm: ProtocolMessage):
    print(f'received: {pm}')
    print(f'message: {repr(pm)}\n')
    return ProtocolMessage.from_bytes(bytes(pm))


def main():
    """Do something to load a seed for private key creation."""
    seed = token_bytes()
    skey = SigningKey(seed)

    """Initialize the session."""
    session = SigningSession({'skey': skey, 'number_of_participants': 2})

    """Create and add the values needed for this participant."""
    n = Nonce()
    nc = NonceCommitment.create(n)
    session.add_nonce_commitment(nc, skey.verify_key)
    session.add_nonce(n, skey.verify_key)

    """Simulate the other guy."""
    other_skey = SigningKey(token_bytes())
    pkey = PublicKey.create([skey.verify_key, other_skey.verify_key])
    other_n = Nonce()
    other_nc = NonceCommitment.create(other_n)
    agg_n = n + other_n
    other_ps = PartialSignature.create(other_skey, other_n.r, pkey.L, pkey.public(),
        agg_n.R, b'obviously a bitcoin transaction')

    """At this point, the state will be AWAITING_PARTICIPANT_KEYS.
        This can be communicated via use of the ProtocolMessage class, which will
        be demonstrated just for this step. Assume there are some methods, send
        and receive, that handle message exchange for us. I am excluding the
        receiving and parsing of ACK messages (and related resending of un-ACKed
        messages) for brevity.
    """
    pm = ProtocolMessage.create(session.id, session.protocol_state, [])
    pm = ProtocolMessage.create(session.id, ProtocolState.SENDING_PARTICIPANT_KEY, [other_skey.verify_key])
    pm = receive(pm.add_signature(other_skey))

    """Parse the message and add the key if possible."""
    try:
        if not pm.check_signature():
            raise Exception('broken signature')

        pm.parse_message()
        if pm.state is ProtocolState.SENDING_PARTICIPANT_KEY:
            vkey = pm.message_parts[0]
            session.add_participant_keys([vkey])
            """Acknowledge receipt of the key (optional)."""
            pm = ProtocolMessage.create(session.id, ProtocolState.ACK_PARTICIPANT_KEY, [vkey])
            send(pm.add_signature(session.skey))
    except ProtocolError as err:
        """Abort on a ProtocolError, communicating the error to peers first."""
        pm = ProtocolMessage.create(session.id, session.protocol_state, [err])
        send(pm.add_signature(session.skey))
        raise err
    except:
        """Retry or something here."""

    """State is now AWAITING_COMMITMENT, so send this participant's commitment."""
    pm = ProtocolMessage.create(session.id, ProtocolState.SENDING_COMMITMENT, [nc])
    send(pm.add_signature(session.skey))

    """Get the vkeys for missing nonce commitments and request those commitments."""
    vkeys = [vk for vk in session.vkeys if vk not in session.nonce_commitments.keys()]
    pm = ProtocolMessage.create(session.id, session.protocol_state, vkeys)
    send(pm.add_signature(session.skey))

    """Receive the commitment."""
    pm = receive(ProtocolMessage.create(
            session.id, ProtocolState.SENDING_COMMITMENT, [other_nc]
        ).add_signature(other_skey))
    try:
        pm.parse_message()
        if pm.state is ProtocolState.SENDING_COMMITMENT:
            nc = pm.message_parts[0]
            session.add_nonce_commitment(nc, pm.vkey)
            pm = ProtocolMessage.create(session.id, ProtocolState.ACK_COMMITMENT, [nc])
            send(pm.add_signature(session.skey))
    except ProtocolError as err:
        """Abort on a ProtocolError, communicating the error to peers first."""
        pm = ProtocolMessage.create(session.id, session.protocol_state, [err])
        send(pm.add_signature(session.skey))
        raise err
    except:
        """Retry or something here."""

    """At this point, I will exclude the message-passing simulation for brevity."""

    """State is now AWAITING_MESSAGE. Once the message is agreed upon, we advance."""
    session.message = b'obviously a bitcoin transaction'

    """State is now AWAITING_NONCE. Get the nonce point from the other participant.
        Once all nonce points are gathered, the aggregate nonce will be calculated.
    """
    try:
        session.add_nonce(other_n, other_skey.verify_key)
    except ProtocolError:
        """If the nonce does not match the commitment, if the vkey is not from a
            known participant, or if too much time has passed while waiting, an
            error will be raised, so deal with that here.
        """

    """State is now AWAITING_PARTIAL_SIGNATURE, so send ours and get theirs.
        Once all partial signatures are gathered, the aggregate signature will
        be calculated.
    """
    ps = session.make_partial_signature()
    session.add_partial_signature(ps, session.skey.verify_key)
    # send ps, receive received_ps and received_vkey
    try:
        session.add_partial_signature(other_ps, other_skey.verify_key)
    except ProtocolError as err:
        """If the received_vkey is not a known participant vkey, or if a
            conflicting partial signature is added when one already exists, an
            handle the error here.
        """

    """State should now be COMPLETE, so distribute the signature."""
    signature = session.signature
    assert session.public_key.verify(signature)
    print(f'{signature=}')
    print(f'{str(signature)=}')

    # compatible with ordinary Ed25519 signature verification
    sig = SignedMessage(bytes(signature))
    vkey = VerifyKey(bytes(session.public_key.public()))
    assert vkey.verify(sig) == session.message


if __name__ == '__main__':
    main()



from context import musig
from json import dumps, loads
from nacl.signing import SigningKey, VerifyKey
import inspect
import unittest


class TestMuSigPublicKey(unittest.TestCase):
    """Test suite for PublicKey."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]
        cls.gvkey = 'eecb56e70d2405a849aa5e55b6e2f96aac2957dba72f8c289994c842e33ec477'

    def test_PublicKey_is_a_class(self):
        assert inspect.isclass(musig.PublicKey)

    def test_PublicKey_init_raises_ValueError_without_proper_arg(self):
        with self.assertRaises(ValueError) as err:
            musig.PublicKey()
        assert str(err.exception) == 'cannot instantiate empty PublicKey'

        with self.assertRaises(TypeError) as err:
            musig.PublicKey((1,2))
        assert str(err.exception) == 'input data must be dict'

    def test_PublicKey_instances_have_key_set_encodings_when_created_from_keys(self):
        aggkey = musig.PublicKey.create(self.verify_keys)
        assert hasattr(aggkey, 'L')
        assert type(aggkey.L) is bytes

    def test_PublicKey_instances_have_gvkey_when_created_from_keys(self):
        aggkey = musig.PublicKey.create(self.verify_keys)
        assert hasattr(aggkey, 'gvkey')
        assert type(aggkey.gvkey) is bytes

    def test_PublicKey_instances_have_vkeys_when_created_from_keys(self):
        aggkey = musig.PublicKey.create(self.verify_keys)
        assert hasattr(aggkey, 'vkeys')
        assert type(aggkey.vkeys) is tuple
        for vk in aggkey.vkeys:
            assert isinstance(vk, VerifyKey)

    def test_PublicKey_instances_have_correct_key_set_encodings(self):
        aggkey1 = musig.PublicKey.create(self.verify_keys[:1])
        aggkey2 = musig.PublicKey.create(self.verify_keys[:2])
        aggkey3 = musig.PublicKey.create(self.verify_keys[:3])

        expected1 = musig.H_small(*[vk.__bytes__() for vk in self.verify_keys[:1]])
        expected2 = musig.H_small(*sorted([vk.__bytes__() for vk in self.verify_keys[:2]]))
        expected3 = musig.H_small(*sorted([vk.__bytes__() for vk in self.verify_keys[:3]]))

        assert aggkey1.L == expected1
        assert aggkey2.L == expected2
        assert aggkey3.L == expected3

    def test_PublicKey_instances_have_correct_aggregate_keys(self):
        aggkey1 = musig.PublicKey.create(self.verify_keys[:1])
        aggkey2 = musig.PublicKey.create(self.verify_keys[:2])
        aggkey3 = musig.PublicKey.create(self.verify_keys[:3])

        expected1 = '03bbdee1f53985e5889bf704a1abe1b313ee302a1630af5fe6cc9f350c94b7e6'
        expected2 = '4c42a4d7403669d652139aa43620ea43b65f8405b10ba34de90a641b86802747'
        expected3 = 'eecb56e70d2405a849aa5e55b6e2f96aac2957dba72f8c289994c842e33ec477'

        assert aggkey1.gvkey.hex() == expected1
        assert aggkey2.gvkey.hex() == expected2
        assert aggkey3.gvkey.hex() == expected3

    def test_PublicKey_from_bytes_raises_ValueError_or_TypeError_when_given_invalid_serialization(self):
        with self.assertRaises(TypeError) as err:
            musig.PublicKey.from_bytes('not bytes')
        assert str(err.exception) == 'cannot call from_bytes with non-bytes param'
        with self.assertRaises(ValueError) as err:
            musig.PublicKey.from_bytes(b'incorrect length')
        assert str(err.exception) == 'byte length must be a multiple of 32'

    def test_PublicKey_from_str_works_when_given_hex_str_with_gvkey(self):
        aggkey = musig.PublicKey.from_str('cc607fdd093041cdd29408d3a26098490b6eba370e00b5808e8c344f47897251')
        assert type(aggkey) is musig.PublicKey
        assert aggkey.gvkey.hex() == 'cc607fdd093041cdd29408d3a26098490b6eba370e00b5808e8c344f47897251'

    def test_PublicKey_from_str_works_when_given_hex_str_of_vkeys(self):
        vkeys = [
            '3139a8eacf6b00b9d420831381a75a73cbef3ffbdaa7796861a034a479844071',
            'da9fe36cb2845734cf1b4c3487340b4f13c519310434893cfa84e275a10f5cac',
            'fc01e0f7c9222529a8f1d5ae3e973745eb163f521e6e347e8f1b14fe9fc15692'
        ]
        aggkey = musig.PublicKey.from_str(''.join(vkeys))
        assert type(aggkey) is musig.PublicKey
        assert aggkey.gvkey.hex() == 'eecb56e70d2405a849aa5e55b6e2f96aac2957dba72f8c289994c842e33ec477'

    def test_PublicKey_public_method_returns_instance_with_only_gvkey(self):
        aggkey = musig.PublicKey.create(self.verify_keys)
        pubkey = aggkey.public()
        assert len(aggkey.vkeys) == 3
        assert len(pubkey.vkeys) == 0
        assert type(pubkey.gvkey) is bytes
        assert pubkey.gvkey == aggkey.gvkey
        assert len(pubkey.vkeys) == 0
        assert pubkey.L is None

    def test_PublicKey_instance_serialize_and_deserialize_properly(self):
        aggkey0 = musig.PublicKey.create(self.verify_keys)
        aggkey00 = aggkey0.public()
        bts1 = bytes(aggkey0)
        bts11 = bytes(aggkey00)
        str2 = str(aggkey0)
        str22 = str(aggkey00)
        js3 = dumps(aggkey0)
        js33 = dumps(aggkey00)

        aggkey1 = musig.PublicKey.from_bytes(bts1)
        aggkey11 = musig.PublicKey.from_bytes(bts11)
        aggkey2 = musig.PublicKey.from_str(str2)
        aggkey22 = musig.PublicKey.from_str(str22)
        aggkey3 = musig.PublicKey(loads(js3))
        aggkey33 = musig.PublicKey(loads(js33))

        assert aggkey0 == aggkey1
        assert aggkey0 == aggkey2
        assert aggkey0 == aggkey3

        assert aggkey00 == aggkey11
        assert aggkey00 == aggkey22
        assert aggkey00 == aggkey33

        assert aggkey0.L is not None
        assert aggkey00.L is None
        assert len(aggkey0.vkeys) > 0
        assert len(aggkey00.vkeys) == 0

    def test_PublicKey_instances_have_verify_method(self):
        aggkey = musig.PublicKey.create(self.verify_keys)
        assert hasattr(aggkey, 'verify')
        assert inspect.ismethod(aggkey.verify)


if __name__ == '__main__':
    unittest.main()



from context import musig
from enum import Enum
from json import dumps, loads
from nacl.signing import SigningKey, VerifyKey, SignedMessage as NaclSignedMessage
from uuid import UUID, uuid4
import inspect
import musig
import unittest


class TestMuSigProtocol(unittest.TestCase):
    """Test suite for ProtocolState, ProtocolError, and ProtocolMessage."""
    @classmethod
    def setUpClass(cls):
        cls.seeds = [
            'bc66e048abf92e97c35f00607a9260dd8299d91e698253c1090872d7d441df80',
            'a7a4b3a2afae8026fb6d523f06f67e5e69ca8e583881ca34574a8e6a9658eaec',
            'a5f496e55953105c5f80939f7a7794edcfd89997e801b6365effd35af1150b02'
        ]
        cls.seeds = [bytes.fromhex(seed) for seed in cls.seeds]
        cls.signing_keys = [SigningKey(seed) for seed in cls.seeds]
        cls.verify_keys = [sk.verify_key for sk in cls.signing_keys]
        cls.gvkey = 'eecb56e70d2405a849aa5e55b6e2f96aac2957dba72f8c289994c842e33ec477'
        cls.uuid = uuid4()

    # ProtocolState tests
    def test_ProtocolState_is_Enum(self):
        assert hasattr(musig, 'ProtocolState')
        assert inspect.isclass(musig.ProtocolState)
        assert issubclass(musig.ProtocolState, Enum)

    def test_ProtocolState_values_are_int(self):
        for attr in musig.ProtocolState:
            assert type(attr.value) is int

    # ProtocolError tests
    def test_ProtocolError_is_class_that_inherits_from_Exception(self):
        assert hasattr(musig, 'ProtocolError')
        assert inspect.isclass(musig.ProtocolError)
        assert issubclass(musig.ProtocolError, Exception)

    def test_ProtocolError_serializes_to_and_from_bytes_properly(self):
        for state in musig.ProtocolState:
            err0 = musig.ProtocolError('important information', state)
            err1 = musig.ProtocolError.from_bytes(bytes(err0))
            assert err0 == err1

    def test_ProtocolError_serializes_to_and_from_str_properly(self):
        for state in musig.ProtocolState:
            err0 = musig.ProtocolError('important information', state)
            err1 = musig.ProtocolError.from_str(str(err0))
            assert err0 == err1

    # ProtocolMessage tests
    def test_ProtocolMessage_is_a_class(self):
        assert hasattr(musig, 'ProtocolMessage')
        assert inspect.isclass(musig.ProtocolMessage)

    def test_ProtocolMessage_init_raises_ValueError_for_empty_or_incorrect_input(self):
        with self.assertRaises(ValueError):
            musig.ProtocolMessage()
        with self.assertRaises(ValueError):
            musig.ProtocolMessage({'state':'', 'not-message':''})
        with self.assertRaises(ValueError):
            musig.ProtocolMessage({'not-state':'', 'message':''})

    def test_ProtocolMessage_init_raises_TypeError_for_nondict_input(self):
        with self.assertRaises(TypeError):
            musig.ProtocolMessage('not a dict')
        with self.assertRaises(TypeError):
            musig.ProtocolMessage(b'not a dict')
        with self.assertRaises(TypeError):
            musig.ProtocolMessage(['not a dict'])

    def test_ProtocolMessage_create_raises_errors_for_improper_params(self):
        with self.assertRaises(TypeError):
            pm = musig.ProtocolMessage.create()
        with self.assertRaises(TypeError):
            pm = musig.ProtocolMessage.create('not a UUID', 'not a ProtocolState', 'not a list')
        with self.assertRaises(TypeError):
            pm = musig.ProtocolMessage.create('not a UUID', musig.ProtocolState.INITIALIZED, [])
        with self.assertRaises(TypeError):
            pm = musig.ProtocolMessage.create(self.uuid, 'not a ProtocolState', [])

    def test_ProtocolMessage_for_EMPTY(self):
        pm = musig.ProtocolMessage.create(None, musig.ProtocolState.EMPTY, [])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.EMPTY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''
        assert hasattr(pm, 'session_id') and pm.session_id is None

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert type(serialized) is bytes
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.EMPTY
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert deserialized.message == b''
        assert hasattr(deserialized, 'session_id') and deserialized.session_id is None

        # check deserialized message has valid signature
        assert deserialized.check_signature()

        # test json serialization and deserialization
        serialized = dumps(deserialized)
        deserialized = musig.ProtocolMessage(loads(serialized))
        assert deserialized.check_signature()
        assert bytes(deserialized) == bytes(pm)

    def test_ProtocolMessage_for_INITIALIZED(self):
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.INITIALIZED, [])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.INITIALIZED
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_AWAITING_PARTICIPANT_KEY(self):
        # scenario: self.signing_keys[0] is awaiting keys
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_PARTICIPANT_KEY, [])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_PARTICIPANT_KEY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_SENDING_PARTICIPANT_KEY(self):
        # scenario: self.signing_keys[0] is sending its vkey
        vkey = self.signing_keys[0].verify_key
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.SENDING_PARTICIPANT_KEY, [vkey])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.SENDING_PARTICIPANT_KEY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(vkey)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert deserialized.message == bytes(vkey)

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_ACK_PARTICIPANT_KEY(self):
        # scenario: self.signing_keys[0] is acknowledging receipt of a vkey
        vkey = self.signing_keys[1].verify_key
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_PARTICIPANT_KEY, [vkey])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_PARTICIPANT_KEY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(vkey)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_keys[0] is acknowledging receipt of 2 vkeys
        vkeys = self.verify_keys[1:]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_PARTICIPANT_KEY, vkeys)
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_PARTICIPANT_KEY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(vk) for vk in vkeys])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert deserialized.message == b''.join([bytes(vk) for vk in vkeys])

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_REJECT_PARTICIPANT_KEY(self):
        # scenario: self.signing_keys[0] is rejecting a vkey
        vkey = self.signing_keys[1].verify_key
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_PARTICIPANT_KEY, [vkey])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_PARTICIPANT_KEY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(vkey)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_keys[0] is rejecting 2 vkeys
        vkeys = self.verify_keys[1:]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_PARTICIPANT_KEY, vkeys)
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_PARTICIPANT_KEY
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(vk) for vk in vkeys])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_AWAITING_COMMITMENT(self):
        # scenario: self.signing_keys[0] is awaiting nonce commitments from the other 2 participants
        vkeys = self.verify_keys[1:]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_COMMITMENT, vkeys)
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_COMMITMENT
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(vk) for vk in vkeys])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_ACK_COMMITMENT(self):
        # scenario: self.signing_keys[0] is acknowledging receipt of a nonce commitment
        nc1 = musig.NonceCommitment.create(musig.Nonce())
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_COMMITMENT, [nc1])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_COMMITMENT
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(nc1)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_keys[0] is acknowledging receipt of 2 nonce commitments
        nc2 = musig.NonceCommitment.create(musig.Nonce())
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_COMMITMENT, [nc1, nc2])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_COMMITMENT
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(nc1), bytes(nc2)])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.ACK_COMMITMENT
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert deserialized.message == b''.join([bytes(nc1), bytes(nc2)])
        assert hasattr(deserialized, 'session_id')
        assert isinstance(deserialized.session_id, UUID) and deserialized.session_id == self.uuid

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_REJECT_COMMITMENT(self):
        # scenario: self.signing_keys[0] is rejecting a received nonce commitment
        nc1 = musig.NonceCommitment.create(musig.Nonce())
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_COMMITMENT, [nc1])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_COMMITMENT
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(nc1)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_keys[0] is rejecting 2 received nonce commitments
        nc2 = musig.NonceCommitment.create(musig.Nonce())
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_COMMITMENT, [nc1, nc2])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_COMMITMENT
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(nc1), bytes(nc2)])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_SENDING_COMMITMENT(self):
        nc = musig.NonceCommitment.create(musig.Nonce())
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.SENDING_COMMITMENT, [nc])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.SENDING_COMMITMENT
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(nc)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_AWAITING_MESSAGE(self):
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_MESSAGE, [])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_MESSAGE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_ACK_MESSAGE(self):
        msg = b'hello world'
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_MESSAGE, [msg])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_MESSAGE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == msg
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.ACK_MESSAGE
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert deserialized.message == msg
        assert hasattr(deserialized, 'session_id')
        assert isinstance(deserialized.session_id, UUID) and deserialized.session_id == self.uuid

    def test_ProtocolMessage_for_REJECT_MESSAGE(self):
        msg = b'hello world'
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_MESSAGE, [msg])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_MESSAGE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == msg
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_SENDING_MESSAGE(self):
        msg = b'hello world'
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.SENDING_MESSAGE, [msg])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.SENDING_MESSAGE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == msg
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_AWAITING_NONCE(self):
        # scenario: self.signing_keys[0] is awaiting nonces for other participants
        vkeys = self.verify_keys[1:]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_NONCE, vkeys)
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_NONCE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(vk) for vk in vkeys])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_keys[0] is awaiting nonces for self.signing_keys[1]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_NONCE, [self.signing_keys[1].verify_key])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_NONCE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(self.signing_keys[1].verify_key)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_ACK_NONCE(self):
        # scenario: self.signing_keys[0] is acknowledging receipt of a nonce
        n1 = musig.Nonce().public()
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_NONCE, [n1])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_NONCE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(n1)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_keys[0] is acknowledging receipt of nonces from others
        n2 = musig.Nonce().public()
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_NONCE, [n1, n2])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_NONCE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(n1) + bytes(n2)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test bytes serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.ACK_NONCE
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert deserialized.message == bytes(n1) + bytes(n2)
        assert hasattr(deserialized, 'session_id')
        assert isinstance(deserialized.session_id, UUID) and deserialized.session_id == self.uuid

        # test str serialization and deserialization
        serialized = str(pm)
        deserialized = musig.ProtocolMessage.from_str(serialized)
        assert deserialized == pm

        # test json serialization and deserialization
        serialized = dumps(pm)
        deserialized = musig.ProtocolMessage(loads(serialized))
        assert deserialized == pm

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_REJECT_NONCE(self):
        n = musig.Nonce().public()
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_NONCE, [n])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_NONCE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(n)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_SENDING_NONCE(self):
        n = musig.Nonce().public()
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.SENDING_NONCE, [n])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.SENDING_NONCE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(n)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_AWAITING_PARTIAL_SIGNATURE(self):
        # scenario: self.signing_key[0] is awaiting partial sigs from other participants
        vkeys = self.verify_keys[1:]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE, vkeys)
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == b''.join([bytes(vk) for vk in vkeys])
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # scenario: self.signing_key[0] is awaiting partial sigs from self.signing_key[1]
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE, [self.signing_keys[1].verify_key])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.AWAITING_PARTIAL_SIGNATURE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(self.signing_keys[1].verify_key)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_ACK_PARTIAL_SIGNATURE(self):
        # scenario: self.signing_keys[0] is acknowledging receipt of a partial signature
        nonce = musig.Nonce()
        pkey = musig.PublicKey.create(self.verify_keys)
        skey = self.signing_keys[1]
        sig = musig.PartialSignature.create(skey, nonce.r, pkey.L, pkey, nonce.R, b'hello world')
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ACK_PARTIAL_SIGNATURE, [sig])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ACK_PARTIAL_SIGNATURE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(sig.public())
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_REJECT_PARTIAL_SIGNATURE(self):
        # scenario: self.signing_keys[0] is acknowledging receipt of a partial signature
        nonce = musig.Nonce()
        pkey = musig.PublicKey.create(self.verify_keys)
        skey = self.signing_keys[1]
        sig = musig.PartialSignature.create(skey, nonce.r, pkey.L, pkey, nonce.R, b'hello world')
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.REJECT_PARTIAL_SIGNATURE, [sig])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.REJECT_PARTIAL_SIGNATURE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(sig.public())
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.REJECT_PARTIAL_SIGNATURE
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert deserialized.message == bytes(sig.public())
        assert hasattr(deserialized, 'session_id')
        assert isinstance(deserialized.session_id, UUID) and deserialized.session_id == self.uuid

        # test str serialization and deserialization
        serialized = str(pm)
        deserialized = musig.ProtocolMessage.from_str(serialized)
        assert deserialized == pm

        # test json serialization and deserialization
        serialized = dumps(pm)
        deserialized = musig.ProtocolMessage(loads(serialized))
        assert deserialized == pm

        # check deserialized message has valid signature
        assert deserialized.check_signature()

    def test_ProtocolMessage_for_SENDING_PARTIAL_SIGNATURE(self):
        # scenario: self.signing_keys[0] is sending a partial sig
        nonce = musig.Nonce()
        pkey = musig.PublicKey.create(self.verify_keys)
        skey = self.signing_keys[0]
        sig = musig.PartialSignature.create(skey, nonce.r, pkey.L, pkey, nonce.R, b'hello world')
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.SENDING_PARTIAL_SIGNATURE, [sig])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.SENDING_PARTIAL_SIGNATURE
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(sig.public())
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

    def test_ProtocolMessage_for_COMPLETED(self):
        ssk = musig.SingleSigKey({'skey': self.signing_keys[0]})
        sig = ssk.sign_message(b'hello world')
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.COMPLETED, [sig])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.COMPLETED
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(sig)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.COMPLETED
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert pm.message == bytes(sig)
        assert hasattr(deserialized, 'session_id')
        assert isinstance(deserialized.session_id, UUID) and deserialized.session_id == self.uuid

    def test_ProtocolMessage_for_ABORTED(self):
        nc = musig.NonceCommitment.create(musig.Nonce().public())
        err = musig.ProtocolError(f'invalid nonce commitment: {nc}', musig.ProtocolState.REJECT_COMMITMENT)
        pm = musig.ProtocolMessage.create(self.uuid, musig.ProtocolState.ABORTED, [err])
        assert isinstance(pm, musig.ProtocolMessage)
        assert hasattr(pm, 'state') and pm.state is musig.ProtocolState.ABORTED
        assert hasattr(pm, 'message') and type(pm.message) is bytes
        assert pm.message == bytes(err)
        assert hasattr(pm, 'session_id')
        assert isinstance(pm.session_id, UUID) and pm.session_id == self.uuid

        # add a signature
        pm.add_signature(self.signing_keys[0])
        assert hasattr(pm, 'signature') and type(pm.signature) is NaclSignedMessage
        assert hasattr(pm, 'vkey') and type(pm.vkey) is VerifyKey

        # test serialization and deserialization
        serialized = bytes(pm)
        assert len(serialized) > 0
        deserialized = musig.ProtocolMessage.from_bytes(serialized)
        assert isinstance(deserialized, musig.ProtocolMessage)
        assert deserialized == pm
        assert hasattr(deserialized, 'state') and deserialized.state is musig.ProtocolState.ABORTED
        assert hasattr(deserialized, 'message') and type(deserialized.message) is bytes
        assert pm.message == bytes(err)
        assert hasattr(deserialized, 'session_id')
        assert isinstance(deserialized.session_id, UUID) and deserialized.session_id == self.uuid

        # check deserialized message has valid signature
        assert deserialized.check_signature()


if __name__ == '__main__':
    unittest.main()
